from typing import Dict
from copy import copy
from datetime import datetime

from WindPy import w

from vnpy.trader.gateway import BaseGateway
from vnpy.trader.object import TickData, SubscribeRequest
from vnpy.trader.utility import ZoneInfo
from vnpy.trader.constant import Exchange


CHINA_TZ = ZoneInfo("Asia/Shanghai")


EXCHANGE_WIND_VT: Dict[str, Exchange] = {
    "SH": Exchange.SSE,
    "SZ": Exchange.SZSE,
    "CFE": Exchange.CFFEX,
    "SHF": Exchange.SHFE,
    "CZC": Exchange.CZCE,
    "DCE": Exchange.DCE
}
EXCHANGE_VT_WIND = {v: k for k, v in EXCHANGE_WIND_VT.items()}


class WindMdApi:
    """Wind行情API实例"""

    wsq_param_map = {}
    wsq_param_map["rt_last"] = "last_price"
    wsq_param_map["rt_last_vol"] = "volume"
    wsq_param_map["rt_oi"] = "open_interest"
    wsq_param_map["rt_open"] = "open_price"
    wsq_param_map["rt_high"] = "high_price"
    wsq_param_map["rt_low"] = "low_price"
    wsq_param_map["rt_pre_close"] = "pre_close"
    wsq_param_map["rt_high_limit"] = "upper_limit"
    wsq_param_map["rt_low_limit"] = "lower_limit"
    wsq_param_map["rt_bid1"] = "bid_price_1"
    wsq_param_map["rt_bid2"] = "bid_price_2"
    wsq_param_map["rt_bid3"] = "bid_price_3"
    wsq_param_map["rt_bid4"] = "bid_price_4"
    wsq_param_map["rt_bid5"] = "bid_price_5"
    wsq_param_map["rt_ask1"] = "ask_price_1"
    wsq_param_map["rt_ask2"] = "ask_price_2"
    wsq_param_map["rt_ask3"] = "ask_price_3"
    wsq_param_map["rt_ask4"] = "ask_price_4"
    wsq_param_map["rt_ask5"] = "ask_price_5"
    wsq_param_map["rt_bsize1"] = "bid_volume_1"
    wsq_param_map["rt_bsize2"] = "bid_volume_2"
    wsq_param_map["rt_bsize3"] = "bid_volume_3"
    wsq_param_map["rt_bsize4"] = "bid_volume_4"
    wsq_param_map["rt_bsize5"] = "bid_volume_5"
    wsq_param_map["rt_asize1"] = "ask_volume_1"
    wsq_param_map["rt_asize2"] = "ask_volume_2"
    wsq_param_map["rt_asize3"] = "ask_volume_3"
    wsq_param_map["rt_asize4"] = "ask_volume_4"
    wsq_param_map["rt_asize5"] = "ask_volume_5"

    def __init__(self, gateway: BaseGateway):
        """"""
        self.gateway: BaseGateway = gateway
        self.gateway_name: str = gateway.gateway_name

        self.connected: bool = False
        self.ticks: Dict[str, TickData] = {}
        self.subscribed: Dict[str, SubscribeRequest] = {}

    def connect(self):
        """"""
        data = w.start()
        if not data.ErrorCode:
            self.gateway.write_log("Wind行情接口连接成功")
        else:
            self.gateway.write_log("Wind行情接口连接失败")
            return

        for req in self.subscribed.values():
            self.subscribe(req)

    def subscribe(self, req: SubscribeRequest):
        """"""
        self.subscribed[req.vt_symbol] = req

        wind_exchange = EXCHANGE_VT_WIND[req.exchange]
        wind_symbol = f"{req.symbol}.{wind_exchange}"

        # 这里额外做一次检查，避免w丢失初始化状态
        if not w.isconnected():
            w.start()

        w.wsq(
            wind_symbol, list(self.wsq_param_map.keys()), func=self.wsq_callback
        )

        self.ticks[wind_symbol] = TickData(
            symbol=req.symbol,
            exchange=req.exchange,
            datetime=datetime.now(),
            gateway_name=self.gateway_name
        )

    def wsq_callback(self, data):
        """"""
        wind_symbol = data.Codes[0]
        tick = self.ticks[wind_symbol]

        dt = data.Times[0]
        tick.datetime = dt.replace(tzinfo=CHINA_TZ)

        for n, field in enumerate(data.Fields):
            field = field.lower()
            key = self.wsq_param_map[field]
            value = data.Data[n][0]
            setattr(tick, key, value)

        self.gateway.on_tick(copy(tick))



from datetime import datetime
from typing import List, Dict, Optional
from math import isnan

from WindPy import w

from vnpy.trader.constant import Exchange, Interval
from vnpy.trader.object import BarData, HistoryRequest
from vnpy.trader.datafeed import BaseDatafeed
from vnpy.trader.utility import ZoneInfo


CHINA_TZ = ZoneInfo("Asia/Shanghai")


EXCHANGE_MAP: Dict[Exchange, str] = {
    Exchange.SSE: "SH",
    Exchange.SZSE: "SZ",
    Exchange.CFFEX: "CFE",
    Exchange.SHFE: "SHF",
    Exchange.CZCE: "CZC",
    Exchange.DCE: "DCE",
}


INTERVAL_MAP: Dict[Interval, str] = {
    Interval.MINUTE: "1",
    Interval.HOUR: "60"
}


class WindDatafeed(BaseDatafeed):
    """万得数据服务接口"""

    def init(self) -> bool:
        """初始化"""
        if w.isconnected():
            return True

        data: w.WindData = w.start()
        if data.ErrorCode:
            return False

        return True

    def query_bar_history(self, req: HistoryRequest) -> Optional[List[BarData]]:
        """查询K线数据"""
        if not w.isconnected():
            self.init()

        if req.interval == Interval.DAILY:
            return self.query_daily_bar_history(req)
        else:
            return self.query_intraday_bar_history(req)

    def query_intraday_bar_history(self, req: HistoryRequest) -> Optional[List[BarData]]:
        """查询日内K线数据"""
        # 参数转换
        wind_exchange: str = EXCHANGE_MAP[req.exchange]
        wind_symbol: str = f"{req.symbol}.{wind_exchange}"

        fields: List[str] = [
            "open",
            "high",
            "low",
            "close",
            "volume",
            "amt",
            "oi"
        ]

        wind_interval: str = INTERVAL_MAP[req.interval]
        options: str = f"BarSize={wind_interval}"

        # 发起查询
        error, df = w.wsi(
            codes=wind_symbol,
            fields=fields,
            beginTime=req.start,
            endTime=req.end,
            options=options,
            usedf=True
        )

        # 检查错误
        if error:
            return []

        # 补全缺失数值
        df.fillna(value=0, inplace=True)

        # 解析数据
        bars: List[BarData] = []
        for tp in df.itertuples():
            dt: datetime = tp.Index.to_pydatetime()

            # 检查是否有持仓量字段
            if isnan(tp.position):
                open_interest: int = 0
            else:
                open_interest: int = tp.position

            bar: BarData = BarData(
                symbol=req.symbol,
                exchange=req.exchange,
                interval=req.interval,
                datetime=dt.replace(tzinfo=CHINA_TZ),
                open_price=tp.open,
                high_price=tp.high,
                low_price=tp.low,
                close_price=tp.close,
                volume=tp.volume,
                turnover=tp.amount,
                open_interest=open_interest,
                gateway_name="WIND"
            )
            bars.append(bar)

        return bars

    def query_daily_bar_history(self, req: HistoryRequest) -> Optional[List[BarData]]:
        """查询日K线数据"""
        # 参数转换
        wind_exchange: str = EXCHANGE_MAP[req.exchange]
        wind_symbol: str = f"{req.symbol}.{wind_exchange}"

        fields: List[str] = [
            "open",
            "high",
            "low",
            "close",
            "volume",
            "amt",
            "oi"
        ]

        # 发起查询
        error, df = w.wsd(
            codes=wind_symbol,
            fields=fields,
            beginTime=req.start,
            endTime=req.end,
            options="",
            usedf=True
        )

        # 检查错误
        if error:
            return []

        # 解析数据
        bars: List[BarData] = []
        for tp in df.itertuples():
            dt: datetime = datetime.combine(tp.Index, datetime.min.time())

            # 检查是否有持仓量字段
            if isnan(tp.OI):
                open_interest: int = 0
            else:
                open_interest: int = tp.OI

            bar: BarData = BarData(
                symbol=req.symbol,
                exchange=req.exchange,
                interval=req.interval,
                datetime=dt.replace(tzinfo=CHINA_TZ),
                open_price=tp.OPEN,
                high_price=tp.HIGH,
                low_price=tp.LOW,
                close_price=tp.CLOSE,
                volume=tp.VOLUME,
                turnover=tp.AMT,
                open_interest=open_interest,
                gateway_name="WIND"
            )
            bars.append(bar)

        return bars
